@@ -19,19 +19,27 @@ std::string Logger::formatString(const char *format, ...)
   return std::string(msg);
 }
 
-Logger::Logger(const char *name)
+void Logger::setupPyLogger()
 {
   py::UniqueRef pyLoggingModule;
-
   py::importPythonModule("snowflake.connector.snow_logging", pyLoggingModule);
   PyObject *logger =
-      PyObject_CallMethod(pyLoggingModule.get(), "getSnowLogger", "s", name);
+      PyObject_CallMethod(pyLoggingModule.get(), "getSnowLogger", "s", m_name);
 
   m_pyLogger.reset(logger);
 }
 
+Logger::Logger(const char *name)
+: m_name(name)
+{
+}
+
 void Logger::log(int level, const char *path_name, const char *func_name, int line_num, const char *msg)
 {
+  if (m_pyLogger.get() == nullptr)
+  {
+    setupPyLogger();
+  }
 
   PyObject *logger = m_pyLogger.get();
   py::UniqueRef keywords(PyDict_New());
