@@ -6,14 +6,13 @@
 
 import logging
 from logging import getLogger
-from typing import Dict
+from typing import TYPE_CHECKING, Dict
 
 from snowflake.connector.constants import UTF8
 
 from .compat import BASE_EXCEPTION_CLASS
 
-MYPY = False
-if MYPY:  # from typing import TYPE_CHECKING once 3.5 is deprecated
+if TYPE_CHECKING:  # pragma: no cover
     from .connection import SnowflakeConnection
     from .cursor import SnowflakeCursor
 
@@ -295,5 +294,5 @@ class OtherHTTPRetryableError(Error):
 class MissingDependencyError(Error):
     """Exception for missing extras dependencies."""
 
-    def __init__(self, dependency):
+    def __init__(self, dependency: str):
         super(MissingDependencyError, self).__init__(msg="Missing optional dependency: {}".format(dependency))
